#include "uwbgl_D3DGraphicsSystem.h"
#include "uwbgl_D3DCollider.h"
#include "uwbgl_D3DDrawablePrimitive.h"
#include "uwbgl_D3DDeviceVertexFormat.h"
#include "uwbgl_D3DFileTexture.h"

using namespace UWBGL_D3D;

ImplementCompoundPrimitiveAttr(DrawablePrimitive, D3DXVECTOR3, Velocity)
ImplementCompoundPrimitiveAttr(DrawablePrimitive, D3DXCOLOR, DrawColor)
ImplementCompoundPrimitiveAttr(DrawablePrimitive, Material, Material)
ImplementSimplePrimitiveAttr(DrawablePrimitive, int, TextureIndex)
ImplementBoolPrimitiveAttr(DrawablePrimitive, Blending)
ImplementBoolPrimitiveAttr(DrawablePrimitive, Texture)
ImplementBoolPrimitiveAttr(DrawablePrimitive, Lighting)


DrawablePrimitive::~DrawablePrimitive()
{
}

DrawablePrimitive::DrawablePrimitive()
: m_Velocity(0.0f, 0.0f, 0.0f)
, m_DrawColor(0.3f, 0.2f, 0.2f, 0.0f)
, m_Texture(false)
, m_TextureIndex(-1)
, m_Blending(false)
, m_Lighting(true)
{
}

bool DrawablePrimitive::OutOfBound(float minx, float miny, float maxx, float maxy)
{
    // returns true if primitive is out of bound, default behavior is to
    // test if vertex[0] is out of bound. If there are no vertices
    // returns false;
  if (NumVertices() == 0) return false;

    return (  (m_Vertices[0].x < minx)  ||  (m_Vertices[0].x > maxx) ||
              (m_Vertices[0].y > maxy)  ||  (m_Vertices[0].y < miny) );
}

void DrawablePrimitive::SetPrimitiveAttributesForDrawing() 
{
  LPDIRECT3DDEVICE9 device = GraphicsSystem::GetSystem().GetActiveDevice();
	if (device== NULL)
		return;
  device->SetFVF(DeviceVertexFormatCode);

	bool textureOn = false;
	if ( m_Texture && (m_TextureIndex >= 0) ) {
			ResourceTable *t = GraphicsSystem::GetSystem().GetActiveResourceTable();
			if (	m_TextureIndex < int(t->size()) ) {
					textureOn = true;
					FileTexture *f = (FileTexture *) ((*t)[m_TextureIndex]);
          if (f != NULL)
					  f->ActivateTexture();
			} 
	}

	if (!textureOn) {
				device->SetTexture(0, NULL);
				device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	}

	if (BlendingIsEnabled()) {
			// 1. enable the blending
			device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);

			// 2. define what we want to do ...
			//
			//   OurColor * (AlphaFromTexture) + (FrameBufferColor * (1-AlphaFromTexture) )
			//
			device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
			device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
	} else {
			device->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	}

  if ( LightingIsEnabled() ) {
      device->SetRenderState(D3DRS_LIGHTING, TRUE);
      device->SetMaterial( m_Material.GetD3DMaterial() );
  } else {
      device->SetRenderState(D3DRS_LIGHTING, FALSE);
  }
} 

//
// Empty functions meant for subclass to define attributes 
// to over-write these functions
//
void DrawablePrimitive::SetPointSize(const float &s) {}
float& DrawablePrimitive::GetPointSize() { static float f(0.0f); return f; }

void DrawablePrimitive::SetFillMode(const WORD&) {}
WORD& DrawablePrimitive::GetFillMode() {static WORD w =D3DFILL_SOLID; return w;}

void DrawablePrimitive::SetShadedMode(const WORD&) {}
WORD& DrawablePrimitive::GetShadedMode() { static WORD w = D3DSHADE_GOURAUD; return w; }

//
// A little real work here ...
//
void
DrawablePrimitive::MovePrimitiveByVelocity()
{
	for (unsigned int i=0; i<m_Vertices.size(); i++)
		D3DXVec3Add(&(m_Vertices[i]),  &(m_Vertices[i]),  &m_Velocity);
}
void DrawablePrimitive::GetPrimitiveCentroid(D3DXVECTOR3 &center)
{
	center = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	for (unsigned int i=0; i<m_Vertices.size(); i++)
		center += m_Vertices[i];
	center.x /= float(m_Vertices.size());
	center.y /= float(m_Vertices.size());
	center.z /= float(m_Vertices.size());
}
void DrawablePrimitive::GetColliderList(ColliderList &list)
{
	float radius = 0.01f;
	D3DXVECTOR3 center;
  
  GetPrimitiveCentroid(center);
	for (unsigned int i=0; i<m_Vertices.size(); i++) {
		D3DXVECTOR3 d = m_Vertices[i] - center;
		float len = D3DXVec3LengthSq(&d);
		if (len > radius)
			radius = len;
	}

	Collider c(center, sqrtf(radius));
	list.Append(c);
}

void DrawablePrimitive::DrawPrimitive()
{
	if ( ShouldDrawPrimitive() ) {

		SetPrimitiveAttributesForDrawing();

		DrawPrimitiveGeometry();

		if ( ShowCollidersIsEnabled() )
			DrawColliders();
			
	}
}

//
// Support for the VertexList ....
const D3DVertexList& DrawablePrimitive::GetVertices() const
{
  return m_Vertices;
}
int DrawablePrimitive::NumVertices() const
{
	return int(m_Vertices.size());
}
void DrawablePrimitive::SetFirstVertex(const D3DXVECTOR3 &v)
{
	if (NumVertices() == 0)
    AppendVertex(v);
  else
    m_Vertices[0] = v;
}
void DrawablePrimitive::AppendVertex(const D3DXVECTOR3 &v)
{
	m_Vertices.push_back(v);
}

// If valid index, replace vertex at index
// otherwise, replace the last vertex
void DrawablePrimitive::ReplaceVertexAt(int index, const D3DXVECTOR3 &v)
{
	if ( (index >= 0) && (index < NumVertices() ) )
		m_Vertices[index] = v;
	else
		m_Vertices[ NumVertices() - 1 ] = v;
}

// If valid index, return vertex at index
// otherwise, return the last vertex
const D3DXVECTOR3& DrawablePrimitive::GetVertexAt(int index) const
{
	if ( (index >= 0) && (index < NumVertices() ) )
		return m_Vertices[index];
	else
		return m_Vertices[ NumVertices() - 1 ];
}

