#include "uwbgl_D3DViewport.h"
#include "uwbgl_D3DGraphicsSystem.h"
#include "uwbgl_D3DFileTexture.h"

using namespace UWBGL_D3D;

/// Class constructor - initializes class members to NULL, 0
D3DViewport::D3DViewport() :
  m_pD3DDevice(NULL),
  m_BackgroundColor(0)
{
}

/// Destructor releases the LPDirect3DDevice9 object
D3DViewport::~D3DViewport()
{
  if(m_pD3DDevice) m_pD3DDevice->Release();

  ClearResourceTable();
	
	// clear the memory control by <vector> (the links)
  m_ResourceTable.clear();
}

/// Returns the state (true/false) of our LPDirect3DDevice9 object
bool D3DViewport::GraphicsDeviceReady()
{
  return (m_pD3DDevice != NULL);
}

/// Returns true on successful connection to graphics device
bool D3DViewport::CreateGraphicsDevice(HWND hWindow)
{
  /// if for some reason, the device is already created, release it!
  if (m_pD3DDevice != NULL)
  {
    m_pD3DDevice->Reset(&m_D3Dpp);
    m_pD3DDevice->Release();
    m_pD3DDevice = NULL;
  }

  /// call parent class function to create graphics connection 
  bool success = ViewportBase::CreateGraphicsDevice(hWindow);
  if(!success) return false;

  /// Create connection to a piece of drawable area for D3D specific interactions (output and input).
  /// VERY IMPORTANT -- Be sure to initialize the graphics system before this point in code
  GraphicsSystem& sys = GraphicsSystem::GetSystem();
  
  /// Get the current desktop display mode, for the default graphics card. If you have more than
  /// one graphics cards on your machine, you may need to do something different here.
  D3DDISPLAYMODE d3ddm;
  if( FAILED( sys.m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) )
  {
    ::MessageBox(hWindow,"D3D Creation Failed during GetAdapterDisplayMode","D3DRenderer::Initialize",MB_OK);
    return false;
  }

  /// Set up the structure used to create the D3dDevice. Most parameters are
  /// zeroed out. We set Windowed to TRUE, since we want to do D3d in a
  /// window, and then set the SwapEffect to "discard", which is the most
  /// efficient method of presenting the back buffer to the display.  And 
  /// we request a back buffer format that matches the current desktop display 
  /// format.
  /// 1. Init the entire D3DPRESENT structure to 0
  ZeroMemory( &m_D3Dpp, sizeof(m_D3Dpp) );
  // 2. We are creating for one window
  m_D3Dpp.Windowed = TRUE;
  // 3. We will discard the buffer content after being swapped back (most efficient)
  m_D3Dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
  m_D3Dpp.BackBufferFormat = d3ddm.Format;
  // 4. Enable Z-buffer stenciling to support 3D ..
  m_D3Dpp.EnableAutoDepthStencil = TRUE;
 	m_D3Dpp.AutoDepthStencilFormat = D3DFMT_D16;

  // Create the Direct3D device. Here we are using the default adapter (most
  // systems only have one, unless they have multiple graphics hardware cards
  // installed) and requesting the HAL (which is saying we want the hardware
  // device rather than a software one). Software vertex processing is 
  // specified since we know it will work on all cards. On cards that support 
  // hardware vertex processing, though, we would see a big performance gain 
  // by specifying hardware vertex processing.

  // notice the somewhat mysterious "hWindow" parameter. This is the MFC-CWnd parameter
  // remember Td3dDrawable is a subclass from CWnd, so, we are using instance variable
  // from our parent class. m_hWnd is the "handler" to "Window" (abstract connection to
  // a window), think now we are creating a D3D-drawable area, that is connected to a MFC-Window
  if( FAILED( sys.m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,  hWindow,
                                        D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                        &m_D3Dpp, &m_pD3DDevice ) ) )
  {
    ::MessageBox(hWindow,"D3D Creation Failed during CreateDevice","D3DRenderer::Initialize",MB_OK);
    return false;
  }
	
  // Setup render states
  // To support lighting and applying illumination ...
  m_pD3DDevice->SetRenderState( D3DRS_LIGHTING,	TRUE);  
  m_pD3DDevice->SetRenderState( D3DRS_SPECULARENABLE, TRUE);
  m_pD3DDevice->SetRenderState( D3DRS_AMBIENT, 0x00101010);  // Ambient light, lights everything in our environment to 16/255 grey.
  m_pD3DDevice->SetRenderState( D3DRS_LOCALVIEWER, TRUE);
  m_pD3DDevice->SetRenderState( D3DRS_NORMALIZENORMALS, TRUE);

  // all Emissive/Ambient/Diffuse/Specular colors used in the lighting computation will come from materials
  m_pD3DDevice->SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_MATERIAL);
  m_pD3DDevice->SetRenderState(D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_MATERIAL);
  m_pD3DDevice->SetRenderState(D3DRS_SPECULARMATERIALSOURCE, D3DMCS_MATERIAL);
  m_pD3DDevice->SetRenderState(D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL);

  m_pD3DDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_FLAT);
  // this says linearly interpolate color from vertex to vertex
  // instead of D3DSHADE_GOURAUD, later we will ...

  m_pD3DDevice->SetRenderState( D3DRS_ZENABLE,	D3DZB_TRUE);
  // Now we are working in 3D, switch on z-buffer
  //   *** NOTE: this is only useful if we specified EnableAutoDepthStencil
  //             when creating the D3DDevice

  m_pD3DDevice->SetRenderState( D3DRS_CULLMODE,	D3DCULL_NONE );
  // Here Says we want to cull all polys with vertices specified in
  // the clockwise order.
  // Other two options are: D3DCULL_CCW  or D3DCULL_NONE
  // where CCW stands for Counter-ClockWise order

  m_pD3DDevice->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);

  // Now set up the system matrices
  D3DXMATRIX identity;
  D3DXMatrixIdentity(&identity);

  // when all three matrices are set to identity, D3D sees a normalized device space (-1 to +1)
  // with origin in the middle of the screen. Now, we will initialize all these for now ...
  m_pD3DDevice->SetTransform(D3DTS_PROJECTION, &identity);
  m_pD3DDevice->SetTransform(D3DTS_VIEW, &identity);
  m_pD3DDevice->SetTransform(D3DTS_WORLD, &identity);

  // Load file textures
  InitializeResourceTable();
  return true;
}


void D3DViewport::SetBackgroundColor(float red, float green, float blue)
{
  m_BackgroundColor = D3DCOLOR_COLORVALUE(red, green, blue, 0.0);
  // the four parameters are R, G, B, and A respectively with 0 represent nothing and
  // 1 represent max. e.g. (0,0,0,X) is black, and (1,1,1,X) is white (X- is don't care)
  // don't worry about the A (alpha-channel) for now, we will learn about this
  // later-on in the quarter, may be in the 9th or 10-th week.
}


void D3DViewport::InitializeDeviceForDrawing()
{
  GraphicsSystem::GetSystem().SetActiveDevice(m_pD3DDevice);
  GraphicsSystem::GetSystem().SetActiveResourceTable(&m_ResourceTable);
}

void D3DViewport::RestoreDeviceAfterDrawing()
{
  GraphicsSystem::GetSystem().SetActiveDevice(NULL);
}
