// UW Bothell Graphics Library Copyright (c) 2004 University of Washington Bothell
// Authors: Kelvin Sung, Steve Baer
// The accompanying library supports CSS Graphics courses taught at UW-Bothell
// See: http://courses.washington.edu/css450/
//      http://courses.washington.edu/css451/
/////////////////////////////////////////////////////////////////////////////////////////

/// uwbgl_MFCUIWindow.cpp: implementation file for uwbgl_MFCUIWindow.h
///
#include "uwbgl_MFC.h"
#include "uwbgl_MFCUIWindow.h"

// DEBUG_NEW macro allows MFC applications to determine memory leak locations in debug builds
#ifdef _DEBUG
  #define new DEBUG_NEW
#endif

using namespace UWBGL_MFC;

// This is a standard MFC macro for defining handler functions for different windows messages
BEGIN_MESSAGE_MAP(UWBGL_MFC::UIWindow, CWnd)
  ON_WM_ERASEBKGND()
  ON_WM_PAINT()
  ON_WM_SIZE()
  ON_WM_LBUTTONDOWN()
  ON_WM_MBUTTONDOWN()
  ON_WM_RBUTTONDOWN()
  ON_WM_MOUSEMOVE()
  ON_WM_LBUTTONUP()
  ON_WM_MBUTTONUP()
  ON_WM_RBUTTONUP()
END_MESSAGE_MAP()

UIWindow::UIWindow(bool timerBased) :
  m_bTimerBased(timerBased),
  m_pViewport(NULL)
{
}

UIWindow::~UIWindow()
{
}

bool UIWindow::ReplaceDialogControl(HWND hDialogHandle, int id)
{
  //check for valid input parameters
  CWnd* pDialog = CWnd::FromHandle(hDialogHandle);
  ASSERT(pDialog!=NULL);
  CWnd* pCurrentCtrl = pDialog->GetDlgItem(id);
  if(NULL==pCurrentCtrl) return false;

  //Copy the contol's styles. This gives us some way to define things like border style
  //in the resource editor
  DWORD style = pCurrentCtrl->GetStyle();
  DWORD styleEx = pCurrentCtrl->GetExStyle();
  // use the control to figure out how large is the initial on-screen area for the graphics window
  CRect rect;
  pCurrentCtrl->GetWindowRect(&rect);

  pCurrentCtrl->DestroyWindow();  // Since the static was just a place holder, destroy it
                                  // convert the static's screen rectangle to the dialog's 
                                  // client coordinates
  
  
  // use the on-screen area information to create our UIWindow
  pDialog->ScreenToClient(&rect);
  BOOL success = CreateEx(styleEx,
                          NULL,     //CWnd default
                          NULL,     //has no name
	                  style|WS_CHILD|WS_CLIPSIBLINGS|WS_CLIPCHILDREN|WS_VISIBLE,
	                  rect,     // Dimension and location of the place-holder
	                  pDialog,  //this is the parent
	                  id);      //Use the ID of the just destroyed static

  BringWindowToTop();
  if(!success) ::MessageBox(hDialogHandle,
                            "UIWindow creations failed!",
                            "UIWindow::ReplaceDialogControl",
                            MB_OK|MB_ICONSTOP);
  return (success==TRUE);
}

/// This just overrides the default window behavior of erasing the background of window. If we
/// don't override this behavior, the window will procude a "blinking" effect.
BOOL UIWindow::OnEraseBkgnd(CDC* pDC)
{
  if(m_pViewport!=NULL) return TRUE;
  return CWnd::OnEraseBkgnd(pDC);
}

/// Handler function for a WM_PAINT message. This is where standard windows draw themselves.
/// Most CSS450/451 windows will draw during a timer event, so we typically won't draw anything
/// at this time.
void UIWindow::OnPaint()
{
  CPaintDC dc(this); // device context for painting
  if(m_pViewport==NULL)
  {
    //If we haven't set up a viewport, draw a message that no viewport has been attached.
    CRect rect;
    GetWindowRect(&rect);
    CRect logRect(rect);
    this->ScreenToClient(&logRect);
    CBrush brush;
    brush.CreateSolidBrush(RGB(0,0,0));
    CBrush* pOldBrush = dc.SelectObject(&brush);
    dc.FillRect(&logRect,&brush);
    CPen pen(PS_SOLID,1,RGB(255,255,255));
    CPen* pOldPen = dc.SelectObject(&pen);
    dc.MoveTo(0,0);
    dc.LineTo(rect.Width(),rect.Height());
    dc.MoveTo(rect.Width(),0);
    dc.LineTo(0,rect.Height());
    this->ScreenToClient(&rect);
    dc.SelectObject(pOldBrush);
    dc.SelectObject(pOldPen);
    dc.DrawText(_T("No Render Viewport Attached"),&rect,DT_CENTER|DT_VCENTER|DT_SINGLELINE);
  }
  else
  {
    //You could optionally set up a renderer to draw on standard window's paint messages.
    //Most renderers in CSS450/451 are going to be timer based for animation purposes.
    if(!m_bTimerBased)
    {
      m_pViewport->InitializeDeviceForDrawing();
      m_pViewport->DrawToDevice();
      m_pViewport->RestoreDeviceAfterDrawing();
    }
  }
}

/// Draw our contained viewport
void UIWindow::DrawViewport()
{
  if(::IsWindow(m_hWnd))
  {
    if(m_pViewport==NULL)
    {
      Invalidate();
    } else {
      m_pViewport->InitializeDeviceForDrawing();
      m_pViewport->DrawToDevice();
      m_pViewport->RestoreDeviceAfterDrawing();
    }
  }
}

/// Handler for a WM_SIZE event. This just resizes the viewport to the same size as the window.
void UIWindow::OnSize(UINT nType, int cx, int cy)
{
  CWnd::OnSize(nType, cx, cy);
  if(m_pViewport) m_pViewport->ResizeGraphicsDevice(cx,cy);
}

/// Left mouse button down handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnLButtonDown(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    //Tell windows that this window should get all mouse messages until capture is released
    SetCapture();
    m_pViewport->OnMouseButton(true,nFlags,point.x,point.y);
  }
}

/// Middle mouse button down handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnMButtonDown(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    //Tell windows that this window should get all mouse messages until capture is released
    SetCapture();
    m_pViewport->OnMouseButton(true,nFlags,point.x,point.y);
  }
}

/// Right mouse button down handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnRButtonDown(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    //Tell windows that this window should get all mouse messages until capture is released
    SetCapture();
    m_pViewport->OnMouseButton(true,nFlags,point.x,point.y);
  }
}

/// Mouse movde handler. Calls OnMouseMove virtual function of the contained viewport.
void UIWindow::OnMouseMove(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    m_pViewport->OnMouseMove(nFlags,point.x,point.y);
  }
}

/// Left mouse button up handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnLButtonUp(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    nFlags |= MK_LBUTTON;
    ::ReleaseCapture();
    m_pViewport->OnMouseButton(false,nFlags,point.x,point.y);
  }
}
/// Middle mouse button up handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnMButtonUp(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    nFlags |= MK_MBUTTON;
    ::ReleaseCapture();
    m_pViewport->OnMouseButton(false,nFlags,point.x,point.y);
  }
}
/// Right mouse button up handler. Calls OnMouseButton virtual function of the contained viewport.
void UIWindow::OnRButtonUp(UINT nFlags, CPoint point)
{
  if(m_pViewport)
  {
    nFlags |= MK_RBUTTON;
    ::ReleaseCapture();
    m_pViewport->OnMouseButton(false,nFlags,point.x,point.y);
  }
}
/// Attach our viewport contained in the window. IViewport parameter calls CreateGraphicsDevice for drawing to viewport.
bool UIWindow::AttachViewport(IViewport* pViewport)
{
  if(pViewport==m_pViewport) return true;

  bool attachSuccess = pViewport->CreateGraphicsDevice(m_hWnd);
  if(attachSuccess) m_pViewport = pViewport;
  else MessageBox("D3D device initialization failed - Error attaching to window!",
                  "UIWindow::AttachViewport",
                  MB_OK|MB_ICONSTOP);
  return attachSuccess;
}
